/*
(c) mdeboer@iua.upf.es 
*/

#include <math.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* macro to declare code, and also store it in a string for generation */
#define CODE(code) \
void process(void) { code } \
char* codestr(void) { return #code; }

/**** how to use: ****

g++ -O2 main_educ.cxx -o main_educ
./main_educ > main_educ_out
./main_educ gen_educ.c
gcc -O2 gen_educ.c -lm -o gen_educ
./gen_educ > gen_educ_out

**********************

time ./main_educ > /dev/null
time ./gen_educ > /dev/null

**********************

./main_educ | aplay -r 44100 -f FLOAT_LE
./gen_educ | aplay -r 44100 -f FLOAT_LE

*********************/

/* the in[] values of the processing classes or of type source
** when they are accessed as 'float', the connected processing is
** executed (thru operator float) and the connected out value is 
** returned
*/
class source
{
public:
	class processing* p;
	int p_i;
	
	source()
	{
		p = 0;
		p_i = 0;
	}
	
	operator float();
};

/* base processing class */
class processing
{
public:
	source *in;
	float *out;
	float *val;
	
	int n_in;
	int n_out;
	int n_val;
	
	char* name;
		
	processing(char* _name,int _n_in,int _n_out,int _n_val)
	:n_in(_n_in),n_out(_n_out),n_val(_n_val)
	{
		in = new source[n_out];
		out = new float[n_out];
		val = new float[n_val];
		name = _name;
		for (int i = 0;i<n_val;i++) val[i] = 0.0;
	}
	
	void set(int i_val,float v)
	{
		val[i_val] = v;
	}
	virtual void process(void) = 0;
	virtual char* codestr(void) = 0;


	/* generate the struct declarations */
	void generate_decl(FILE* f)
	{
		char* ptr = codestr();
		while (*ptr)
		{
		  /* recurse */
			if (strncmp(ptr,"in[",3)==0)
			{
				ptr+=3;
				int i = atoi(ptr);
				in[i].p->generate_decl(f);
			}
			ptr++;
		}
		fprintf(f,"struct s_%s {\n",name);
		if (n_val) fprintf(f,"float val[%d];\n",n_val);
		if (n_in) fprintf(f,"float in[%d];\n",n_in);
		if (n_out) fprintf(f,"float out[%d];\n",n_out);
		fprintf(f,"} %s;\n\n",name);
	}

	/* generate the val initializations */
	void generate_init(FILE* f)
	{
		char* ptr = codestr();
		while (*ptr)
		{
		  /* recurse */
			if (strncmp(ptr,"in[",3)==0)
			{
				ptr+=3;
				int i = atoi(ptr);
				in[i].p->generate_init(f);
			}
			ptr++;
		}
		for (int i=0;i<n_val;i++)
		{
			fprintf(f,"%s.val[%d] = %f;\n",name,i,val[i]);
		}
	}
	
	/* generate the processing code */
	void generate(FILE* f)
	{
		char* ptr = codestr();
		while (*ptr)
		{
		  /* recurse */
			if (strncmp(ptr,"in[",3)==0)
			{
				ptr+=3;
				int i = atoi(ptr);
				in[i].p->generate(f);
				fprintf(f,"%s.in[%d] = %s.out[%d];\n",name,i,in[i].p->name,in[i].p_i);
			}
			ptr++;
		}
		char tmp[2048];
		char* tgt = tmp;
		ptr = codestr();
		while (*ptr)
		{
			/* turn in, out and val members into addressed struct members*/
			if (
				strncmp(ptr,"in[",3)==0 || 
				strncmp(ptr,"out[",4)==0 ||
				strncmp(ptr,"val[",4)==0 )
			{
				strcpy(tgt,name);
				tgt+=strlen(name);
				*tgt++='.';
			}
			if (*ptr==';')
			{
				*tgt++ = *ptr++;
				*tgt++='\n';
			}else{
				*tgt++ = *ptr++;
			}
		}
		*tgt = 0;
		fprintf(f,tmp);
	}
};

source::operator float()
{
	p->process();
	return p->out[p_i];
}

class var:public processing
{
public:
	float cur;
	var(char* n):processing(n,0,1,1) {}
	CODE(
		out[0] = val[0];
	)
};

class phase:public processing
{
public:
	phase(char* n):processing(n,1,1,1) {}
	CODE(
		out[0] = val[0];
		val[0] = val[0] + in[0];
		if (val[0]>=2.*M_PI) val[0]-=2.*M_PI;
	)
};

class sinus:public processing
{
public:
	sinus(char* n):processing(n,1,1,0) {}
	CODE(
		out[0] = sin(in[0]);
	);
};

class plus:public processing
{
public:
	plus(char* n):processing(n,2,1,0) {}
	CODE(
		out[0] = in[0]+in[1];
	)
};

class mul:public processing
{
public:
	mul(char* n):processing(n,2,1,0) {}
	CODE(
		out[0] = in[0]*in[1];
	)
};

float outbuf[1024];
int outbufi = 0;

class output:public processing
{
public:
	output(char* n):processing(n,1,0,0) {}
	CODE(
		outbuf[outbufi++] = in[0];
	);
};

void connect(processing& source,int i_out,processing& target,int i_in)
{
	target.in[i_in].p = &source;
	target.in[i_in].p_i = i_out;
}

int main(int argc,char** argv)
{
	/* a simple FM model */
	
	var car_dphase("car_dphase");
	var mod_dphase("mod_dphase");
	var mod_index("mod_index");
	
	phase car_phase("car_phase");
	phase mod_phase("mod_phase");
	
	sinus car_sinus("car_sinus");
	sinus mod_sinus("mod_sinus");

	plus my_plus("my_plus");
	mul  my_mul("my_mul");

	output out("output");

	car_dphase.set(0,0.04);
	mod_dphase.set(0,0.08);
	mod_index.set(0,10.);

	connect(mod_dphase,0,mod_phase,0);
	connect(mod_phase,0,mod_sinus,0);
	connect(mod_sinus,0,my_mul,0);
	connect(mod_index,0,my_mul,1);
	connect(my_mul,0,my_plus,0);
	
	connect(car_dphase,0,car_phase,0);
	connect(car_phase,0,my_plus,1);
	
	connect(my_plus,0,car_sinus,0);
	connect(car_sinus,0,out,0);

	if (argc==1)
	{
		/* runtime execution */
		for (int j=0;j<1000;j++)
		{
			outbufi = 0;
			for (int i=0;i<1024;i++)
			{
				out.process();
			}
			write(1,outbuf,1024*sizeof(float));
		}
	}
	else 
	{
		/* code generation */
		FILE* f = fopen(argv[1],"w");
		
		fprintf(f,"#include <math.h>\n");
		fprintf(f,"#include <unistd.h>\n");

		out.generate_decl(f);

		fprintf(f,"main() {\n");
		fprintf(f,"\tfloat outbuf[1024];\tint outbufi;\n\tint j;\n");

		out.generate_init(f);

		fprintf(f,"\tfor (j=0;j<1000;j++) {\n");
		fprintf(f,"\t\toutbufi = 0;\n");
		fprintf(f,"\t\twhile (outbufi<1024) {\n");

		out.generate(f);

		fprintf(f,"\t\t}\n");
		fprintf(f,"\t\twrite(1,outbuf,1024*sizeof(float));\n");
		fprintf(f,"\t}\n");
		fprintf(f,"\treturn 0;\n");
		fprintf(f,"}\n");
		
		fclose(f);
	}
	return 0;
}
